#ifndef DMRPSOLVER_HPP_
#define DMRPSOLVER_HPP_

#include <vector>
#include <utility>
#include <list>
#include "GlobalSettings.hpp"
#include "BCPEngine.hpp"
#include "Decision.hpp"
#include "RestartStrategies.hpp"


class DeltaClause{
	Literal lit;
	const SharedClause* cls;
public:
	DeltaClause():lit(INVALID_LIT),cls(NULL){assert(false);};
	DeltaClause(const SharedClause* cl):lit(cl->getAt(0)),cls(cl){};
	DeltaClause(const DeltaClause& cl):lit(cl.lit),cls(cl.cls){};
	const SharedClause* getClause() const {return cls;}
	Literal	getLiteral() const			{return lit;}
	void	setLiteral(Literal l)		{lit = l;}
};

class OneLevDeltaClses{    
    uint32_t valid 		: 1;
    uint32_t sz			:31;
    uint32_t mem;
    DeltaClause* deltas;
    void newDelta(){deltas = (DeltaClause*)malloc(mem* sizeof(DeltaClause));}
    void ensureOne(){
    	if(sz!=mem)		{asrt(sz<mem);asrt(deltas!=NULL);return;}
		if(deltas==NULL){asrt(sz==0);mem=4; newDelta(); return;}

		DeltaClause* od	= deltas;
		mem*=2;	newDelta();
		memcpy(deltas,od,sz* sizeof(DeltaClause));
		free(od);
    }
 
public:
    OneLevDeltaClses()						:valid(1),sz(0),mem(0),deltas(NULL){}
    bool isValid() const					{return valid;}
    void invalidate()						{if(deltas!=NULL){free(deltas);} valid=sz=mem=0; deltas=NULL;}
    uint32_t size() const					{return sz;}
    uint32_t addClsGetIndex(const SharedClause* cls)	{ensureOne();deltas[sz++]=DeltaClause(cls);return sz-1;}
    DeltaClause* deltasBegin() 				{return deltas;}
    DeltaClause& getAt(uint32_t idx)		{asrt(idx<sz);return deltas[idx];}
};

struct DCPtr{
    uint32_t idx;
    DCPtr(uint32_t i=max32):idx(i){}
};

struct MakeClause{
    uint32_t 	index_in_level;
    DCPtr		level_clses;
    MakeClause(){}
    MakeClause(const uint32_t idx, DCPtr dcls):index_in_level(idx),level_clses(dcls){};
    MakeClause(const MakeClause& cp):index_in_level(cp.index_in_level),level_clses(cp.level_clses){};
};

class DCContainer{
    vector<OneLevDeltaClses> 	all_lev_deltas;
  public:
    DeltaClause& getDC(const MakeClause& mc){asrt(mc.level_clses.idx<all_lev_deltas.size());
      return all_lev_deltas[mc.level_clses.idx].getAt(mc.index_in_level);}
   
    OneLevDeltaClses& getOneLev(const DCPtr& ptr){asrt(ptr.idx<all_lev_deltas.size());
       return all_lev_deltas[ptr.idx];}
    
    OneLevDeltaClses& getOneLev(const MakeClause& mc){return getOneLev(mc.level_clses);}

    void invalidate(const DCPtr& ptr){asrt(ptr.idx<all_lev_deltas.size());
		all_lev_deltas[ptr.idx].invalidate();}

    void clearAll(){all_lev_deltas.clear();}
      
    DCPtr newLevel(){
      all_lev_deltas.push_back(OneLevDeltaClses());
      return DCPtr(all_lev_deltas.size()-1);}      
};







class DMRPTrooper : public BcpEngine<VarAct,TrooperSiblingInfo,SubsumDummy> {
	typedef BcpEngine<VarAct,TrooperSiblingInfo,SubsumDummy> super;
	volatile uint32_t		timeout;
	ConflictsAndGC			conflicts;
	Restarts				restarts;

    DCContainer				all_deltas; //collect all deltas for cleaning
    
    vector<DCPtr>			delta_stack; //a vector of delta clauses for each decision level
    uint32_t 				delta_stack_idx;
	
    BitVector<Variable>		ref_point;
    VecOfLists<ClsPointer,Variable> hook_resp;
    VecOfLists<MakeClause,Variable> var_makes;

    std::list<std::pair<SharedClause*,uint32_t> > //keep copy of lbd to allow sorting (claus' lbd may be changed by other thread)
							unlinked_clses;



    struct MakeCountDescendingFunctor{
	DMRPTrooper* outer ;
	MakeCountDescendingFunctor(DMRPTrooper* s):outer(s){};
	bool operator() (const Literal& a, const Literal& b){
	  return outer->var_makes.sizeOf(a.toVar()) > outer->var_makes.sizeOf(b.toVar());	}
    };

protected:
    bool 	isRefPoint(const Literal& l)const {return ref_point.isSet(l.toVar()) == l.isPos();}
    Literal getRefPoint(const Variable& v)const{return v.toLit(ref_point.isSet(v));}
    void	setRefPoint(const Variable& v, const bool b){b ? ref_point.setFlag(v) : ref_point.clear(v);}
    void	setRefPoint(const Literal& l){setRefPoint(l.toVar(),l.isPos());}

	bool 	relinkUnhookedClauses();
	void	unlinkUnhookedClauses();
	void 	unlinkUnhookedOf(const Literal& lit);
	TrBool	computeNewHooks();

	//void	resetNormalClsState();
	void 	clearAssignsCareForHooks();
	uint32_t findNewHookResponsibles(const Variable& var, vector<std::pair<Literal,ClsPointer> >& link_later);
	Literal findHookRespons(SharedClause* cls, uint32_t& sz_out)const;

	//bool	checkUnwatchedClauses();

	bool	clsLinkHook(ClsPointer cl);
	void 	clsHookNewLearnt(const ClsPointer& clp, const SharedClause* cls, const Literal& ass);
	void 	initHookResponses();


	TrBool  dmrpSolve();
	TrBool 	dmrpSearchSatClause(SharedClause* cls);
	TrBool	tryToLinkAndHook(SharedClause* cls);
	bool	containsHooked(SharedClause* cls);

	uint32_t handleConflictDMRP(int32_t bcp_res);
	void	restoreProbClsDelta();
	void	clearDeltaStack(uint32_t start_idx, const bool del);
	void 	addProbClsDelta(const SharedClause* cls,const uint level);
	inline void incrementMakeCount(const Literal& lit,const  DCPtr& ldcl, const uint32_t idx);
	inline void addClsUnlinked(SharedClause* cls,const bool back);
	void 	registerUnlinkedClausesForMakeCount();

	void 	updateDelta(uint32_t q_begin);
	void 	updateDeltaCheckClause(DeltaClause*& from,DeltaClause*& copy);

	void 	modifyHooksAt0(uint32_t q_begin);
	TrBool  checkAndClearBrokenClauses();
	Literal nextDeltaDecision();
	uint32_t collectPossibleDecLits(DeltaClause& dcls,SharedClause* cls,vector<Literal>& out);
	uint32_t computeMakeCount(const Literal& lit,const uint32_t& lower_bound, const SharedClause* ommit);

	inline void	deltaProbClausesEnsureLevel(uint32_t level);
	void	clearDeltaClauses();
	void	removeRubbishDeltaClauses(const bool clear_global);
	bool	removeRubbishMakeCounts(Variable var);

	/*inline void incDecCount(const Literal& lit);
	inline uint16_t getDecCount(const Literal& lit);*/

	void 	cancelUntilCheck(uint32_t level);
	//int32_t	findNewWatcherConsiderHooks(Clause& cl,const Literal& false_lit, uint32_t other_watcher_idx,int32_t& out_free_idx);
	void 	cancelUntilKeepHooks(uint32_t level);

public:
	inline DMRPTrooper(Satistics* stat,uint32_t to,VarAct* dec, TrooperSiblingInfo* sinf,
			Restarts restart_init,const ConflictsAndGC& confl,const bool glue,const HyperBinRes hbres);

	inline void	stopASAP();

	inline void createClone(const SolvState<SimplyFire,LeaderSiblingInfo>& state, const vector<Literal>& units); //before parallel start
	static inline DMRPTrooper* create(const uint32_t& grp_sz, const uint32_t& t_id, const UserBit& leader_bit,ElimClses& eli,
				uint32_t seed, VarAct::act_t max,VarAct::act_t decay,ImportCriteria imprt,
				uint32_t rand_freq_permil,Restarts restart_init,
				const ConflictsAndGC& confl,uint32_t to,const bool glue_max,
				const HyperBinRes hbres);

	static DMRPTrooper* destruct(DMRPTrooper* s){
				Satistics* st=s->super::sstate.stats;	TrooperSiblingInfo* sinf=s->super::sstate.s_info;
				VarAct* dec	=s->super::decision;
				delete s;  delete dec; delete sinf; delete st; return NULL;}

	TrBool 		isSat();



protected:
	uint32_t asrtGetHookedCount() const ;
	bool asrtAtLeastTwoUnassignedLits(const SharedClause* cl,uint level) const ;
	bool asrtNumberHookResponsiblesAllRight() const;
	bool asrtAllWatchedAlsoHooked()const;
	bool asrtAllWatchedAlsoHookedOf(const Literal& lit,uint32_t & count)const;
	bool asrtAllFalse(const SharedClause* cls,const uint32_t end, const Literal& except);




};



DMRPTrooper::DMRPTrooper(Satistics* stat,uint32_t to,VarAct* dec, TrooperSiblingInfo* sinf,
		Restarts restart_init,const ConflictsAndGC& confl,const bool glue,const HyperBinRes hbres):
	super						(stat,dec,sinf,glue,NoBUP,hbres,NoOnTheFlySub),
	timeout						(to),
	conflicts					(confl),
	restarts					(restart_init),
	delta_stack_idx				(0),
	hook_resp('H'),
	var_makes('M')
{}

DMRPTrooper* DMRPTrooper::create(const uint32_t& grp_sz, const uint32_t& t_id, const UserBit& leader_bit,ElimClses& eli,
			uint32_t seed, VarAct::act_t max,VarAct::act_t decay,ImportCriteria imprt,
			uint32_t rand_freq_permil,Restarts restart_init,
			const ConflictsAndGC& confl,uint32_t to,const bool glue,
			const HyperBinRes hbres){

	Satistics* stats 				= new Satistics();
	TrooperSiblingInfo* sinf		= new TrooperSiblingInfo(stats,grp_sz,t_id,leader_bit,imprt,eli);
	VarAct* dec						= new VarAct(stats,max,decay,seed,rand_freq_permil,False);
	DMRPTrooper* ret 				=
			new DMRPTrooper(stats,to,dec,sinf,restart_init,confl,glue,hbres);
	sinf->initVarInfo(&ret->sstate.vinf);
	return ret;
}




void DMRPTrooper::createClone(const SolvState<SimplyFire,LeaderSiblingInfo>&  state, const vector<Literal>& units){
	super::createClone(state,units);
	conflicts			.setNumberClauses(super::sstate.stats->nAllClauses());
	const uint32_t vars	= nVars();
	ref_point			.ensureSize(vars,false);
	hook_resp			.enlarge(vars);
	var_makes			.enlarge(vars);

	LOG3("Unit clauses finally inited");
	//dec_var_count	.resize(numberVariables(),0);
}



void	DMRPTrooper::stopASAP(){
	super::sstate.stats->nbr_confl = conflicts.count();
	timeout = 0;
	restarts.stopASAP();
}




void	DMRPTrooper::deltaProbClausesEnsureLevel(uint32_t level){
	delta_stack_idx = level + 1;

	if(delta_stack.size() >= delta_stack_idx ){
		delta_stack.resize(delta_stack_idx);
	}else{
		for(uint i = delta_stack.size();i < delta_stack_idx;++i){
			delta_stack.push_back(all_deltas.newLevel());
		}
	}
	asrt(all_deltas.getOneLev(delta_stack[level]).isValid());
}

void DMRPTrooper::incrementMakeCount(const Literal& lit,const DCPtr& ldcl, const uint32_t idx){
	asrt(!isRefPoint(lit));
	var_makes.add(lit.toVar(),MakeClause(idx,ldcl));
}

void DMRPTrooper::addClsUnlinked(SharedClause* cls,const bool back){
	std::pair< SharedClause*,uint32_t> pr(cls,cls->getCount());
	back ? unlinked_clses.push_back(pr): unlinked_clses.push_front(pr);
}


/*
void DMRPTrooper::incDecCount(const Literal& lit){
	uint16_t& count = dec_var_count[lit.toVar().idx()];
	if(count == MAX16) {
		print_out("var reached max \n");
		for(uint i = dec_var_count.size();i--;){dec_var_count[i] >>= 6;}
	}
	++count;
}

uint16_t DMRPTrooper::getDecCount(const Literal& lit){
	return dec_var_count[lit.toVar().idx()];
}*/



/*
struct ClauseEqualFunc{
   	const Clause* comp;
   	ClauseEqualFunc(Clause* compare):comp(compare){assert(compare != NULL);};
   	bool operator() (const Clause* e){return e == comp;}
};
*/


const char*  dmrptrooper_cpp_version();
#define dmrptrooper_hpp_version "DMRPSolver.hpp:     $Revision: 1.9 $"



#endif /*HYBRIDSOLVER_HPP_*/
