/*
 * Decision.hpp
 *
 *  Created on: 27.10.2009
 *      Author: stephan64
 */

#ifndef DECISION_HPP_
#define DECISION_HPP_
#define decision_hpp_version      "$Revision: 1.4 $"

#include "GlobalSettings.hpp"
#include "Satistics.hpp"
#include "Random.hpp"
#include "VariableInfos.hpp"
#include "Heap.hpp"


enum PolarityMode{Prev,Rand,False,True,Ignore};

typedef LazyFraction<7,uint16_t,uint16_t> var_act_t;
typedef LazyFraction<7,uint16_t,uint16_t> lit_act_t;

struct VarActComp{
	bool better(const Variable& va, const var_act_t& a,const Variable& vb, const var_act_t& b)const{
		return a > b;}
	VarActComp(){}
};



class VarAct{
public:
	typedef var_act_t 					act_t;
protected:
	BinVarInfo*							v_info;
	Satistics*							stats;
	Random								randy;
	VarActComp							var_act_comp;
	BumpHeap< Variable, var_act_t ,VarActComp>	var_act_hp;

	uint32_t							rand_freq;
	PolarityMode						pol_mode;

	inline Variable decideVarUnrepl();
public:
	uint32_t getRandFreq() const 			{return rand_freq;}
	void ensureSize(const uint32_t sz)		{var_act_hp.reserve(sz);}
	void validateNbrVariables(const uint32_t sz){
		LOG1("seed " << randy.getLastInitSeed());
		var_act_hp.insertElementsTo(sz,var_act_t(0));}
	void litInConflict(const Literal& lit)	{var_act_hp.activityBump(lit.toVar());}
	void forgetVar(const Variable& var)		{var_act_hp.removeIfContained(var);}
	void replVar(const Variable& die, const Variable& surive);
	void restoreVar(const Variable & var)	{var_act_hp.restoreElementIf(var);}
	void decay()							{var_act_hp.decayAll();}
	act_t getAct(const Literal& l)			{return var_act_hp.getValue(l.toVar());}
	act_t getMaxAct()						{return var_act_hp.getMax();}
	act_t getHalfOfMax()					{return var_act_hp.getMax().getHalf();}
	bool timeForRandDecision()				{return randy.nextInt(1000) < rand_freq;}
	uint32_t randInt(const uint32_t n)		{asrt(n>0);return randy.nextInt(n);}

	void setVarInfo(BinVarInfo* v){v_info=v;}

	inline VarAct(Satistics* s,act_t maxa, act_t dec,uint32_t seed,uint32_t freq,PolarityMode m);
	inline Literal decide();
	inline Variable decideVar();
};



struct LitActComp{
	bool better(const Literal& la, const lit_act_t& a,const Literal& lb, const lit_act_t& b)const{
		return a > b;}
	LitActComp(){}
};


class LitAct{
public:
	typedef lit_act_t act_t;
protected:
	BinVarInfo*							v_info;
	Satistics*							stats;
	Random								randy;
	LitActComp							lit_act_comp;
	BumpHeap< Literal, act_t ,LitActComp>  		lit_act_hp;

	uint32_t							rand_freq;

public:
	uint32_t getRandFreq() const 			{return rand_freq;}
	void ensureSize(const uint32_t sz)		{lit_act_hp.reserve(sz*2);}
	void validateNbrVariables(const uint32_t sz){
		LOG1("seed " << randy.getLastInitSeed());
		lit_act_hp.insertElementsTo(2*sz,lit_act_t(0)); } //true if not yet contained
	void litInConflict(const Literal& lit)	{lit_act_hp.activityBump(lit);}
	void forgetVar(const Variable& var)		{lit_act_hp.removeIfContained(var.toNegLit());
											 lit_act_hp.removeIfContained(var.toPosLit());}
	void replVar(const Variable& die, const Variable& surive);
	void restoreVar(const Variable & var)	{lit_act_hp.restoreElementIf(var.toNegLit());
											 lit_act_hp.restoreElementIf(var.toPosLit());}
	void decay()							{lit_act_hp.decayAll();}
	act_t getAct(const Literal& l)			{return lit_act_hp.getValue(l);}
	act_t getMaxAct()						{return lit_act_hp.getMax();}
	act_t getHalfOfMax()					{return lit_act_hp.getMax().getHalf();}
	bool timeForRandDecision()				{return randy.nextInt(1000) < rand_freq;}
	uint32_t randInt(const uint32_t n)		{asrt(n>0);return randy.nextInt(n);}

	void setVarInfo(BinVarInfo* v){v_info=v;}

	inline LitAct(Satistics* s,act_t maxa, act_t dec,uint32_t seed,uint32_t freq,PolarityMode dummy);
	inline Literal decide();
};




VarAct::VarAct(Satistics* s,act_t var_max_act, act_t var_decay,
		uint32_t rand_seed,uint32_t rand_freq_permil,PolarityMode mode)
	:v_info			(NULL),
	 stats			(s),
	 randy			(rand_seed),
	 var_act_hp		(var_max_act,var_decay,var_act_comp),
	 rand_freq		(rand_freq_permil),
	 pol_mode		(mode){}



LitAct::LitAct(Satistics* s,act_t lit_max_act, act_t lit_decay,
		uint32_t rand_seed,uint32_t rand_freq_permil,PolarityMode dummy)
	:v_info			(NULL),
	 stats			(s),
	 randy			(rand_seed),
	 lit_act_hp		(lit_max_act,lit_decay,lit_act_comp),
	 rand_freq		(rand_freq_permil){}
	 //pol_mode		(mode){}






Variable VarAct::decideVarUnrepl(){
	TIME_ON(stats->decision_time);
	Variable var = INVALID_VAR;

	if(timeForRandDecision()){
		//random decision
		++stats->nbr_rand_dec;
		do{
			if(var_act_hp.empty()){
				TIME_OFF(stats->decision_time);
				return INVALID_VAR;}
			var	= var_act_hp.removeRandom(randy);

		}while(v_info->varKnwn(var));

	}else{
		//activity based decision
		++stats->nbr_act_dec;
		do{
			if(var_act_hp.empty()){
				TIME_OFF(stats->decision_time);
				return INVALID_VAR;}
			var	= var_act_hp.removeTop();
		}while(v_info->varKnwn(var));
	}

	asrt(var != INVALID_VAR );
	TIME_OFF(stats->decision_time);
	return var;
}

Variable VarAct::decideVar(){
	Variable var;
	do{
		var = decideVarUnrepl();
		if(var == INVALID_VAR) return var;

		var = v_info->findLitRepl(var.toPosLit()).toVar();

	}while(v_info->varElim(var) || v_info->varKnwn(var));
	return var;
}


Literal VarAct::decide(){

	Literal lit;
	do{
		Variable var = decideVar();
		if(var == INVALID_VAR) return INVALID_LIT;

		switch(pol_mode){
		case Prev	: {lit = v_info->varPrevValWasTrue(var) ? var.toPosLit() : var.toNegLit(); break;}
		case False	: {lit = var.toNegLit(); break;}
		case True	: {lit = var.toPosLit(); break; }
		case Rand	: {lit = randy.nextBool() ?	var.toNegLit():var.toPosLit(); break;}
		default: asrt(false); return var.toNegLit();
		}
		lit = v_info->findLitRepl(lit);
	}while(v_info->litKnwn(lit));

	return lit;

}


Literal LitAct::decide(){
	TIME_ON(stats->decision_time);
	Literal lit 	= INVALID_LIT;
	const bool rand	= timeForRandDecision();
	if(rand) ++stats->nbr_rand_dec; else ++stats->nbr_act_dec;

	do{
		if(lit_act_hp.empty()){
			TIME_OFF(stats->decision_time);
			return INVALID_LIT;}
		lit	= rand?lit_act_hp.removeRandom(randy):lit_act_hp.removeTop();
		lit = v_info->findLitRepl(lit);
	}while(v_info->varElim(lit.toVar()) || v_info->litKnwn(lit));

	asrt(lit != INVALID_LIT );
	TIME_OFF(stats->decision_time);
	return lit;
}



#endif /* DECISION_HPP_ */
