#ifndef SCC_HPP_
#define SCC_HPP_
#define scc_hpp_version "$Revision: 1.2 $"


#include "GlobalSettings.hpp"
#include "BitVector.hpp"
#include "AtomicTypes.hpp"
#include "BCPEngine.hpp"
#include "SiblingInfo.hpp"
#include "SimplyFire.hpp"
#include <list>

using std::list;

class SCC{

	SolvState<SimplyFire,LeaderSiblingInfo>& sstate;

	static const uint64_t MSK_ST		= 3;
	static const uint64_t MSK_IDX		= max32 & ~MSK_ST;
	static const uint64_t MSK_CLO		= max64 ^ max32;
	static const uint64_t MSK_NO_ST		= max64 & ~MSK_ST;
	static const uint64_t MSK_NO_IDX	= max64 & ~MSK_IDX;
	static const uint64_t MSK_NO_CLO	= max64 & ~MSK_CLO;

	bool stateIs(const Literal& l, const uint32_t& v)const{asrt(l.idx()<sstate.cls_reg.lit_space.size());
		return (sstate.cls_reg.lit_space[l.idx()] & MSK_ST) == v;}
	uint32_t getIdx(const Literal& l){
		return (sstate.cls_reg.lit_space[l.idx()] & MSK_IDX ) >>2;}


	void setState(const Literal& l, const uint32_t v){asrt(v<4);
		uint64_t& r = sstate.cls_reg.lit_space[l.idx()]; r = (MSK_NO_ST&r) | v; asrt(stateIs(l,v));}

	void setIdx(const Literal& l,const uint32_t& t){
		uint64_t& r = sstate.cls_reg.lit_space[l.idx()]; r = (MSK_NO_IDX&r) | (t<<2); asrt(getIdx(l)==t);}


	uint32_t ST_CLEAR; //0 OR 2
	uint32_t ST_FINI;//2 OR 0

	bool stClear(const Literal& l) const		{return stateIs(l,ST_CLEAR);}
	bool stStck(const Literal& l) const 		{return stateIs(l,1);}
	bool stFini(const Literal& l) const 		{return stateIs(l,ST_FINI);}
	void doStck(const Literal& l){setState(l,1);}
	void doFini(const Literal& l){setState(l,ST_FINI);}

	uint32_t sccCollIdx(const Literal& l){asrt(stStck(l));return getIdx(l);}
	void sccSetCollIdx(const Literal& l,const uint32_t& t){asrt(stStck(l)); setIdx(l,t);}

	uint32_t rdntClosureIdx(const Literal& l){asrt(l.idx()<sstate.cls_reg.lit_space.size());
		return sstate.cls_reg.lit_space[l.idx()] >> 32;}
	void	 rdntSetClosureIdx(const Literal& l,const uint64_t t){asrt(l.idx()<sstate.cls_reg.lit_space.size());
		uint64_t& r = sstate.cls_reg.lit_space[l.idx()]; r = (MSK_NO_CLO&r) | (t<<32); asrt(rdntClosureIdx(l)==t);}


protected:


	struct SCCCollect{
		Literal 	lit;
		uint32_t 	root; //index into scc_collect where root of this scc starts
		SCCCollect(const Literal& l, const uint32_t& r):lit(l),root(r){}
	};
	vector<SCCCollect>		scc_collect;

	SCCCollect& getCollIfo(const Literal& lit){ asrt(stStck(lit));
		return scc_collect[sccCollIdx(lit)];}

	struct StackEle{
		uint32_t 	coll_top;//index into scc_collect to actual literal
		uint32_t 	child_strt; //index into closure
		uint32_t	next_child;//index into closure

		StackEle(const uint32_t& coll_sz, const uint32_t& clsr_sz)//, const uint32_t& closure)
			:coll_top(coll_sz), child_strt(clsr_sz),next_child(clsr_sz){}//,prev_closure(closure){}

		//const StackEle& operator=(const StackEle& e) const{return *this;}//dummy
	};
	vector<StackEle>			scc_stack;

	SCCCollect& getCollIfoByStackIdx(const uint32_t& stk_idx){
			asrt(stk_idx < scc_stack.size()); asrt(scc_stack[stk_idx].coll_top < scc_collect.size());
				return scc_collect[scc_stack[stk_idx].coll_top];}


	bool checkRoot(const Literal& lit) const{ return sstate.wtch.noBinsPart(lit);}
	bool sccPushStack(const Literal& lit);
	void sccPopCollect(Literal& lit);
	void sccPopStack(const Literal& top,vector<Message>* rem_bin_out);
	void sccStart(const Literal& lit,list<list<Literal> >& sccs,vector<Literal>& units_out, vector<Message>* remv_bins_out);

////remove redundant binaries

	struct Closure{
		uint32_t lit : 31;
		uint32_t vld : 1;
		Closure(const Literal& l, const bool& v):lit(l.idx()),vld(v){}
		Closure():lit(0),vld(0){}
	};
	vector<Closure>	  closure;


public:
	void init(){uint64_t t = max32; t = (t<<32) | ST_FINI;	LOG2("init in scc all with " << t);
		LOG2("Masks: st:" << MSK_ST << " idx:" <<MSK_IDX<< " clo:" <<MSK_CLO<<" no_st:" << MSK_NO_ST << " no_idx:" <<MSK_NO_IDX<< " no_clo:" <<MSK_NO_CLO);
		for(uint32_t i=sstate.cls_reg.lit_space.size();i--;){sstate.cls_reg.lit_space[i] = t;}}

	SCC( SolvState<SimplyFire,LeaderSiblingInfo>& ss)
		:sstate(ss),ST_CLEAR(2),ST_FINI(0){}

	//void computeSCCs();
	void computeSCCs(list<list<Literal> >& sccs_out,vector<Literal>& units_out, vector<Message>* remv_bins_out);

	//bool removeRedundant(vector<Literal>& units_out,vector<Literal>* sinks_out);

	//sanity
private:
	bool asrtAllState(const uint32_t& v) const;
	bool asrtNotReplaced(const Literal& l,const char* why)const;
	bool asrtCheckSanity(const uint32_t min_type) const ;
	bool asrtBinPartExists(const Literal& a, const Literal& b) const;
	//bool asrtClsPartExists(const Literal& lit, const ClsP& clp) const;
	bool asrtFlgsClear(const bool chck_prnt) const;
	bool asrtWatchesCorrect()const;
	bool asrtAllRepl(const list<Literal>& scc) const;
	friend class Settings;
};


const char*  scc_cpp_version();
#endif /* SCC_HPP_ */
