
#include "GlobalSettings.hpp"
#include "SolverFactory.hpp"
#include "SolverLeader.hpp"
#include <signal.h>

using std::cout; using std::endl;

#define NO_INTERRUPT


struct ClsCheck{
	vector<bool> vals;
	IFStream in;
	uint32_t clses;
	bool ok;
	ClsCheck(const string& file):in(file.c_str()),clses(0),ok(true){}
	void reserveForVariables(const uint32_t& i){}
	void reserveForClauses(const uint32_t & i){}
	bool addClause(const vector<Literal>& lits,const bool lrnt){
		++clses;
		for(uint32_t i=lits.size();i--;){
			Literal lit = lits[i]; asrt(lit.toVar().idx()<vals.size());
			if(vals[lit.toVar().idx()] == lit.isPos()){ return true;}
		}
		ok = false;
		ASRT_PRNT("The following clause with size "<<lits.size() <<" is not SAT: " << endl);
		for(uint32_t i=0;i!=lits.size();++i){std::cerr << " " << lits[i].toDIMACS();} std::cerr << " /" << endl;
		return true; //we want to see all false clauses
	}
	void invalidateAll(){ok = false;}
	bool check(){
		if(in.isBad())	{ASRT_PRNT("Could not open CNF file for debug"); return false;}
		parse(in,*this); in.close();
		if(ok){			cout<<"c Verified " << clses << " clauses " << endl;}
		return ok;}
};


int 	SingleSolver::printResult(const UserID& uid,const string& file,TrBool result){
	if(result.isTrue()){
		printf("\ns SATISFIABLE\n");
		solver->printModel();

		/*ClsCheck chk(file);
		solver->getModel(chk.vals);
		if(!chk.check()){cout << "could not verify model" << endl;}*/
		return EXIT_SAT;}

	if(result.isFalse()){
		printf("\ns UNSATISFIABLE\n");
		char* why = solver->super::whyInvalid();
		if(why != NULL){printf("c %s\n",why);}
		return EXIT_UNSAT;}

	printf("\ns UNKNOWN\n");
	return EXIT_UNKNOWN;
}

#ifndef NO_INTERRUPT
SingleSolver* sisolv 		= NULL;
ExplicitSolver* exsolv		= NULL;


static void SIGTERM_handler(int signum) {
	fprintf(stderr,"\n       !!!INTERRUPTED!!!       \n");
	if(sisolv != NULL && sisolv->canPrintFinal()){
		sisolv->stopAllASAP();
		sisolv->printStatsConsole(UserID(0),UNKNOWN_TrBool);
		//sisolv->printDIMACS(UserID(0),"dflt_dimacs.cnf",UNKNOWN_TrBool);
	}

	if(exsolv != NULL && exsolv->canPrintFinal()){
		exsolv->stopAllASAP();
		exsolv->printStatsConsole(UserID(0),UNKNOWN_TrBool);
		//musolv->printDIMACS(UserID(0),"dflt_dimacs.cnf",UNKNOWN_TrBool);
	}

    exit(EXIT_UNKNOWN);
};

static void SIGINT_handler(int signum) {

	fprintf(stderr,"\n       !!!SOME INFOS!!!       \n");
	if(sisolv != NULL){
		sisolv->printStatsConsole(UserID(0),UNKNOWN_TrBool);
	}
	if(exsolv != NULL){
		exsolv->printStatsConsole(UserID(0),UNKNOWN_TrBool);
	}

	fprintf(stderr,"     !!!SOME INFOS - USE KILL TO TERMINATE!!!   \n");
};
#else

static void SIGTERM_handler(int signum) {
	printf("\ns UNKNOWN\n");
    exit(EXIT_UNKNOWN);
};

static void SIGINT_handler(int signum) {
	printf("\ns UNKNOWN\n");
	exit(EXIT_UNKNOWN);
};

#endif

SingleSolver::SingleSolver(GlobalSettings& sets):printed(0){


	ConflictsAndGC confl(25000, 20000,200, 300,0.5);
	solver = solver_t::create(&sets,1,0,elim,decision_t::act_t(1000), decision_t::act_t(135,1), ImportCriteria(256,false,1,1,1),
				20,Prev,Restarts::createLuby(100),confl, true,OptimisticBUP,BiereHyperBin,DoOnTheFlySub);
	LOG3("SOLVER CREATED" << endl);

#ifndef NO_INTERRUPT
	sisolv = this;
#endif
	signal(SIGINT,SIGINT_handler);
	signal(SIGTERM,SIGTERM_handler);
}




bool SingleSolver::addClause(vector<Literal>& literals,const bool lrnt){

	Literal max = solver->validateCls(literals);
	if(max == TRUE_LIT) return true;

	const uint32_t sz = literals.size();
	if(sz == 0 || max == FALSE_LIT) return false;

	asrt(sz == 1 || literals[0] != literals[1]);

	//solver->reserveForVariables(max_var.idx()+1);

	if(sz <= 2){
		return solver->addValidUnsharedClause(literals[0],sz==1?INVALID_LIT:literals[1]);
	}

	SharedClause* cls = SharedClause::construct(solver->sstate.s_info->all_users_bits,	literals.begin(),literals.end(),sz-1,lrnt);
	solver->addValidSharedClause(cls,literals.size());

	return true;
}

void 	SingleSolver::printCSV(const UserID& uid,const std::string& file_name, TrBool result){
		solver->sstate.stats->printCSV(file_name,result);}

void 	SingleSolver::printDIMACS(const UserID& uid,const std::string& file_name, TrBool result){
		solver->printDIMACS(file_name,result);}

void 	SingleSolver::printStatsConsole(const UserID& uid, TrBool result){
		solver->sstate.stats->printStatsOneThread(uid,result,1);}

void 	SingleSolver::checkOut(const UserID& uid){
		asrt(solver!= NULL);
		solver=solver_t::destruct(solver);}

TrBool	SingleSolver::invokeSolver(const UserID& uid){
		return solver->isSat();}

void 	SingleSolver::stopPrinting(){
		solver->print_cons = false;}

void 	SingleSolver::stopAllASAP(){
		solver->stopASAP();}

void 	SingleSolver::reserveForVariables(uint32_t s){
		solver->reserveForVariables(s);}

void 	SingleSolver::reserveForClauses(uint32_t s){
		solver->reserveForClauses(s);}

void 	SingleSolver::invalidateAll(){
		solver->setInvalid("At init SingleSolver");}



/*
 *
 * explicit solver
 *
 *
 * */


#define destructIf(i) if(trooper ## i != NULL){ troop_t ## i ::destruct(trooper ## i);}
#define callIf(i,t) if(trooper ## i != NULL){ trooper ## i ->t;}
#define addIf(i,t,v) if(trooper ## i != NULL){ v.push_back(trooper ## i ->t );}


#define callAll(t) callIf(1,t) callIf(2,t) callIf(3,t) callIf(4,t) callIf(5,t) callIf(6,t) callIf(7,t)
#define addAll(t,v)	addIf(1,t,v) addIf(2,t,v) addIf(3,t,v) addIf(4,t,v) addIf(5,t,v) addIf(6,t,v) addIf(7,t,v)

#define applyChoose(uid,t) \
		switch(uid.id){\
		case 0:{leader->t;break;}\
		case 1:{trooper1->t;break;}\
		case 2:{trooper2->t;break;}\
		case 3:{trooper3->t;break;}\
		case 4:{trooper4->t;break;}\
		case 5:{trooper5->t;break;}\
		case 6:{trooper6->t;break;}\
		case 7:{trooper7->t;break;}}

#define valChoose(uid,t,v) \
		switch(uid.id){\
		case 0:{v=leader->t;break;}\
		case 1:{v=trooper1->t;break;}\
		case 2:{v=trooper2->t;break;}\
		case 3:{v=trooper3->t;break;}\
		case 4:{v=trooper4->t;break;}\
		case 5:{v=trooper5->t;break;}\
		case 6:{v=trooper6->t;break;}\
		case 7:{v=trooper7->t;break;}}


int 	ExplicitSolver::printResult(const UserID& uid,const string& file,TrBool result){

	//printf("\nc Using %d threads",(uint32_t)threads);
	if(result.isTrue()){
		printf("\ns SATISFIABLE\n");


		/*char* why = NULL;
		valChoose(uid,whyInvalid(),why)
		if(why!=NULL){printf("c Strange should be unsat %s\n",why);}*/


		//ClsCheck chk(file);
		//applyChoose(uid,getModel(chk.vals) )

		applyChoose(uid,printModel() )

		/*bool ok;
		valChoose(uid,asrtAllLinkedClsesSAT(chk.vals),ok)
		if(!ok){cout << "solver itself could not verify model" << endl;}
		else{cout << "solver itself verified model" << endl;}
		if(!chk.check()){cout << "could not verify model" << endl;}*/

		return EXIT_SAT;}

	if(result.isFalse()){
		printf("\ns UNSATISFIABLE\n");
		char* why = NULL;
		valChoose(uid,whyInvalid(),why)
		if(why!=NULL){printf("c %s\n",why);}
		return EXIT_UNSAT;}

	printf("\ns UNKNOWN\n");
	return EXIT_UNKNOWN;
}
/*
void 	ExplicitSolver::printDIMACS(const UserID& uid,const string& file,TrBool result){
	if(uid.id == 0)	{lead_solver->printDIMACS(file,result);}
	else{ troop_solvers[uid.id-1]->printDIMACS(file,result);}
}*/



/*
void ExplicitSolver::initTypes(GlobalSettings& sets){
	asrt(threads <= 8); asrt(threads >=2);
	uint32_t id			= 0;

	leader = lead_t::create(&sets,threads,id++,elim,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,true,7,6,40),
					20,Prev,Restarts::createStatic(100),ConflictsAndGC(35000, 30000,200, 300, 0.5),
					true,OptimisticBUP,BiereHyperBin,DoOnTheFlySub);

	const UserBit& leader_bit 	= leader->sstate.s_info->user_bit;
	uint32_t rand_seed 			= sets.init_rand_seed + sets.inc_rand_seed;

	trooper1 = id<threads ? troop_t1::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,false,7,6,50),
					15,Prev,Restarts::createGlucose(150,70), ConflictsAndGC(30000, 20000,500, 300,0.5),sets.timeout,
					true,NoBUP,BiereHyperBin,DoOnTheFlySub): NULL;

	trooper2 = id<threads ? troop_t2::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(139,1),ImportCriteria(1024,true,5,4,50),
					15,False, Restarts::createGeometric(100,1.5), ConflictsAndGC(30000, 20000,500, 300,0.5), sets.timeout,
					false,PessimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;



	trooper3 = id<threads ? troop_t3::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,true,3,2,70),
					15, Restarts::createGlucose(100,70),	ConflictsAndGC(30000, 20000,500, 300,0.5), sets.timeout,
					true,NoHyperBin): NULL;


	trooper4 = id<threads ? troop_t4::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,false,5,4,30),
					20,Prev, Restarts::createLuby(64), ConflictsAndGC(35000, 25000,200, 300,0.5), sets.timeout,
					true,PessimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;



	trooper5 = id<threads ? troop_t5::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					LitAct::act_t(1000), LitAct::act_t(137,1),ImportCriteria(256,false,3,2,70),
					15,Ignore, Restarts::createGeometric(100,1.5),ConflictsAndGC(30000, 20000,500, 300,0.5), sets.timeout,
					false,OptimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;

	trooper6 = id<threads ? troop_t6::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(139,1),ImportCriteria(256,false,5,4,20),
					15,Prev, Restarts::createLuby(32),ConflictsAndGC(30000, 20000,500, 300,0.5), sets.timeout,
					false,NoBUP,BiereHyperBin,DoOnTheFlySub): NULL;


	trooper7 = id<threads ? troop_t7::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(133,1),ImportCriteria(256,true,7,6,50),
					20,Prev, Restarts::createGlucose(100,75),ConflictsAndGC(30000, 20000,500, 300,0.5), sets.timeout,
					true,PessimisticBUP,NoHyperBin,DoOnTheFlySub): NULL;


	linkOthers(trooper1);
	linkOthers(trooper2);
	linkOthers(trooper3);
	linkOthers(trooper4);
	linkOthers(trooper5);
	linkOthers(trooper6);
	linkOthers(trooper7);

	LOG2("SOLVER CREATED" << endl);

#ifndef NO_INTERRUPT
	exsolv = this;
#endif
	signal(SIGINT,SIGINT_handler);

}
*/




void ExplicitSolver::initTypes(GlobalSettings& sets){

	asrt(threads <= 8); asrt(threads >=2);
	uint32_t id			= 0;

	leader = lead_t::create(&sets,threads,id++,elim,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,true,3,2,90),
					20,Prev,Restarts::createStatic(100),ConflictsAndGC(35000, 30000,200, 300, 0.75),
					true,OptimisticBUP,BiereHyperBin,DoOnTheFlySub);

	const UserBit& leader_bit 	= leader->sstate.s_info->user_bit;
	uint32_t rand_seed 			= sets.init_rand_seed + sets.inc_rand_seed;

	trooper1 = id<threads ? troop_t1::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,false,3,2,70),
					20,Prev,Restarts::createGlucose(100,70), ConflictsAndGC(30000, 20000,500, 300,0.5),sets.timeout,
					true,PessimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;

	trooper2 = id<threads ? troop_t2::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(139,1),ImportCriteria(1024,true,3,2,50),
					15,False, Restarts::createGeometric(100,1.5), ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					false,OptimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;



	trooper3 = id<threads ? troop_t3::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,true,3,2,70),
					15, Restarts::createStatic(50),ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					true,NoHyperBin): NULL;


	trooper4 = id<threads ? troop_t4::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,true,7,6,80),
					20,Prev, Restarts::createGlucose(200,70), ConflictsAndGC(35000, 25000,200, 300,0.75), sets.timeout,
					true,OptimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;



	trooper5 = id<threads ? troop_t5::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					LitAct::act_t(1000), LitAct::act_t(137,1),ImportCriteria(256,false,3,2,80),
					15,Ignore, Restarts::createGeometric(100,1.5),ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					false,OptimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;

	trooper6 = id<threads ? troop_t6::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(139,1),ImportCriteria(256,false,5,4,70),
					15,Prev, Restarts::createLuby(16),ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					false,OptimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;


	trooper7 = id<threads ? troop_t7::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(133,1),ImportCriteria(256,true,7,6,80),
					20,Prev, Restarts::createGlucose(100,75),ConflictsAndGC(50000, 40000,1, 1,0.5), sets.timeout,
					true,PessimisticBUP,NoHyperBin,DoOnTheFlySub): NULL;


	linkOthers(trooper1);
	linkOthers(trooper2);
	linkOthers(trooper3);
	linkOthers(trooper4);
	linkOthers(trooper5);
	linkOthers(trooper6);
	linkOthers(trooper7);

	LOG2("SOLVER CREATED" << endl);

#ifndef NO_INTERRUPT
	exsolv = this;
#endif
	signal(SIGINT,SIGINT_handler);

}

/*
void ExplicitSolver::initTypes(GlobalSettings& sets){

	asrt(threads <= 8); asrt(threads >=2);
	uint32_t id			= 0;

	leader = lead_t::create(&sets,threads,id++,elim,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,true,3,2,90),
					20,Prev,Restarts::createStatic(100),ConflictsAndGC(35000, 30000,200, 300, 0.75),
					true,NoBUP,NoHyperBin,NoOnTheFlySub);

	const UserBit& leader_bit 	= leader->sstate.s_info->user_bit;
	uint32_t rand_seed 			= sets.init_rand_seed + sets.inc_rand_seed;

	trooper1 = id<threads ? troop_t1::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
						VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,true,3,2,70),
						15, Restarts::createGlucose(100,70),	ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
						true,NoHyperBin): NULL;

	trooper2 = id<threads ? troop_t2::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(139,1),ImportCriteria(1024,true,3,2,70),
					15,False, Restarts::createGeometric(100,1.5), ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					false,NoBUP,NoHyperBin,NoOnTheFlySub): NULL;



	trooper3 = id<threads ? troop_t3::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,false,3,2,70),
					15,Prev,Restarts::createGlucose(200,90), ConflictsAndGC(50000, 40000,200, 300,0.5),sets.timeout,
					true,NoBUP,NoHyperBin,NoOnTheFlySub): NULL;





	trooper4 = id<threads ? troop_t4::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,false,3,2,80),
					20,Prev, Restarts::createLuby(64), ConflictsAndGC(35000, 25000,200, 300,0.75), sets.timeout,
					true,NoBUP,NoHyperBin,NoOnTheFlySub): NULL;



	trooper5 = id<threads ? troop_t5::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					LitAct::act_t(1000), LitAct::act_t(137,1),ImportCriteria(256,false,3,2,80),
					15,Ignore, Restarts::createGeometric(100,1.5),ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					false,NoBUP,NoHyperBin,NoOnTheFlySub): NULL;

	trooper6 = id<threads ? troop_t6::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(139,1),ImportCriteria(256,false,5,4,70),
					15,Prev, Restarts::createLuby(16),ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					false,NoBUP,NoHyperBin,NoOnTheFlySub): NULL;


	trooper7 = id<threads ? troop_t7::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(133,1),ImportCriteria(256,true,7,6,80),
					20,Prev, Restarts::createGeometric(100,1.3),ConflictsAndGC(50000, 40000,1, 1,0.5), sets.timeout,
					false,NoBUP,NoHyperBin,NoOnTheFlySub): NULL;


	linkOthers(trooper1);
	linkOthers(trooper2);
	linkOthers(trooper3);
	linkOthers(trooper4);
	linkOthers(trooper5);
	linkOthers(trooper6);
	linkOthers(trooper7);

	LOG2("SOLVER CREATED" << endl);

#ifndef NO_INTERRUPT
	exsolv = this;
#endif
	signal(SIGINT,SIGINT_handler);

}
*/

/*

void ExplicitCraftedSolver::initTypes(GlobalSettings& sets){


	asrt(threads <= 8); asrt(threads >=2);
	uint32_t id			= 0;

	leader = lead_t::create(&sets,threads,id++,elim,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,3,2,90),
					20,Prev,Restarts::createStatic(100),ConflictsAndGC(35000, 30000,200, 300, 0.75),
					2,OptimisticBUP,BiereHyperBin,DoOnTheFlySub);

	const UserBit& leader_bit 	= leader->sstate.s_info->user_bit;
	uint32_t rand_seed 			= sets.init_rand_seed + sets.inc_rand_seed;

	trooper1 = id<threads ? troop_t1::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
						VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,3,2,70),
						15,Prev, Restarts::createGeometric(100,1.5), ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
						0,OptimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;

	trooper2 = id<threads ? troop_t2::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
						VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,7,6,80),
						20,Prev, Restarts::createGeometric(100,1.3),ConflictsAndGC(50000, 40000,1, 1,0.5), sets.timeout,
						0,PessimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;

	trooper2 = id<threads ? troop_t2::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,3,2,80),
					20,Prev, Restarts::createNoRestarts(), ConflictsAndGC(50000, 40000,500, 500,0.5), sets.timeout,
					0,PessimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;

	trooper3 = id<threads ? troop_t3::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,3,2,70),
					15, Restarts::createLuby(100),	ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					2,BiereHyperBin): NULL;

	trooper3 = id<threads ? troop_t3::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(135,1),ImportCriteria(256,3,2,70),
					15,Ignore, Restarts::createLuby(100),	ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					2,PessimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;

	trooper4 = id<threads ? troop_t4::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(139,1),ImportCriteria(256,5,4,70),
					15,Prev, Restarts::createLuby(32),ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					0,OptimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;



	trooper5 = id<threads ? troop_t5::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					LitAct::act_t(1000), LitAct::act_t(137,1),ImportCriteria(256,3,2,80),
					15,Ignore, Restarts::createGeometric(100,1.5),ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					0,OptimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;

	trooper6 = id<threads ? troop_t6::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(139,1),ImportCriteria(256,3,2,70),
					15,False, Restarts::createGeometric(100,1.5), ConflictsAndGC(50000, 40000,200, 300,0.5), sets.timeout,
					0,NoBUP,BiereHyperBin,DoOnTheFlySub): NULL;

	trooper7 = id<threads ? troop_t7::create(threads,id++,leader_bit,elim,rand_seed+=sets.inc_rand_seed,
					VarAct::act_t(1000), VarAct::act_t(133,1),ImportCriteria(256,7,6,80),
					20,Prev, Restarts::createGeometric(100,1.3),ConflictsAndGC(50000, 40000,1, 1,0.5), sets.timeout,
					0,PessimisticBUP,BiereHyperBin,DoOnTheFlySub): NULL;


	linkOthers(trooper1);
	linkOthers(trooper2);
	linkOthers(trooper3);
	linkOthers(trooper4);
	linkOthers(trooper5);
	linkOthers(trooper6);
	linkOthers(trooper7);

	LOG2("SOLVER CREATED" << endl);

#ifndef NO_INTERRUPT
	exsolv = this;
#endif
	signal(SIGINT,SIGINT_handler);
}
*/


void 	ExplicitSolver::initSharing(){

	running = leader->sstate.s_info->all_users_bits;
	leader->sstate.s_info->activate();
	callAll(createClone(leader->sstate, leader->assign_stack) )
}

void 	ExplicitSolver::printCSV(const UserID& uid,const std::string& file_name, TrBool result){
	vector<Satistics*> stats;
	stats.push_back(leader->sstate.stats);
	addAll(sstate.stats,stats)
	Satistics::printCSV(file_name,stats,uid,result);
}


void 	ExplicitSolver::printStatsConsole(const UserID& uid, TrBool result){
	vector<Satistics*> stats;
	stats.push_back(leader->sstate.stats);
	addAll(sstate.stats,stats)
	Satistics::printStatsConsole(stats,uid,result);
}

void 	ExplicitSolver::checkOut(const UserID& uid){
	UserBit ub(uid);
	if(!running.removeUser(ub)){
		LOG1("Checkout solver without destruction");
		return;}

	LOG1("Destroy queues content of troopers. I am " << uid.id);
	callAll(sstate.s_info->deleteQueueContent() )
	LOG1("Destroy queues content of leader");
	leader->sstate.s_info->deleteQueueContent();

	LOG1("Destruct troopers");
	destructIf(1) destructIf(2) destructIf(3) destructIf(4)
	destructIf(5) destructIf(6) destructIf(7)
	LOG1("Destruct leader");
	lead_t::destruct(leader);
	LOG1("Everything destructed ");
}

TrBool	ExplicitSolver::invokeSolver(const UserID& uid){
	//UserBit ub(uid);
	//running.addUser(ub);
	TrBool res = UNKNOWN_TrBool;
	valChoose(uid,isSat(),res)
	return res;
}

void 	ExplicitSolver::stopPrinting(){
	leader->print_cons = false;}



void 	ExplicitSolver::stopAllASAP(){
	callAll(stopASAP())
	leader->stopASAP();
}

void 	ExplicitSolver::reserveForVariables(uint32_t s){
	leader->reserveForVariables(s);}

void 	ExplicitSolver::reserveForClauses(uint32_t s){
	leader->reserveForClauses(s);}

void 	ExplicitSolver::invalidateAll(){
	callAll(setInvalid("At init ExplicitSolver"))
	leader->setInvalid("At init ExplicitSolver");
}



bool ExplicitSolver::addClause(vector<Literal>& literals,const bool lrnt){

	Literal max = leader->validateCls(literals);
	if(max == TRUE_LIT) return true;

	const uint32_t sz = literals.size();
	if(sz == 0 || max == FALSE_LIT) return false;

	asrt(sz == 1 || literals[0] != literals[1]);

	if(sz <= 2){
		return leader->addValidUnsharedClause(literals[0],sz==1?INVALID_LIT:literals[1]);
	}

	SharedClause* cls = SharedClause::construct(leader->sstate.s_info->all_users_bits,literals.begin(),literals.end(),sz-1,lrnt);
	leader->addValidSharedClause(cls,literals.size());

	return true;
}






const char*  solverfactory_cpp_version(){return "$Revision: 1.21 $";}

